import './types';
import fastify, { FastifyReply, FastifyRequest } from 'fastify';
import fastifyCookie from '@fastify/cookie';
import fastifyCors from '@fastify/cors';
import mercurius from 'mercurius';
import mercuriusCodegen from 'mercurius-codegen';
import { PrismaClient } from '@prisma/client';
import config from './config';
import { resolvers } from './resolvers';
import schema from './schema';
import { MemcService, MinioService, TokenService } from './services';

const {
  server: { allowedOrigins, ...serverConfig },
  minio: minioConfig,
  token: tokenConfig,
  memc: memcConfig,
  cookie: cookieConfig,
} = config;

const app = fastify({
  logger: true,
});

const prisma = new PrismaClient();
const minio = new MinioService(minioConfig);
const memc = new MemcService(memcConfig);
const token = new TokenService({ ...tokenConfig, memc, prisma });

setInterval(
  () => {
    token.cleanupExpiredTokens().catch((err) => {
      app.log.error('Failed to clean up expired tokens:', err);
    });
  },
  60 * 60 * 1000
); // 1 hour

const context = async (req: FastifyRequest, reply: FastifyReply) => {
  const { authorization } = req.headers;
  const accessToken = authorization?.replace('Bearer ', '');
  const jwt = await token.verifyAccessToken(accessToken);
  await minio.initialize();
  return {
    prisma,
    jwt,
    minio,
    token,
    memc,
    req,
    reply,
  };
};

app.register(fastifyCookie, {
  secret: cookieConfig.secret,
  parseOptions: {
    secure: config.isProduction,
    httpOnly: true,
    sameSite: config.isProduction ? 'none' : 'lax',
    path: '/',
    maxAge: 60 * 60 * 24 * 30, // 30 days
    signed: true,
  },
});

app.register(fastifyCors, {
  origin: (origin, callback) => {
    if (!origin || allowedOrigins.includes(origin)) return callback(null, true);
    return callback(new Error('Not allowed'), false);
  },
  methods: ['GET', 'POST', 'PUT', 'DELETE', 'OPTIONS'],
  credentials: true,
});

app.register(mercurius, {
  schema,
  subscription: true,
  jit: 1,
  cache: true,
  context,
});

app.register(async ({ graphql }) => {
  resolvers.forEach((resolver) => {
    graphql.defineResolvers(resolver);
  });
});

app.get('/ping', async () => {
  return 'pong\n';
});

mercuriusCodegen(app, {
  targetPath: './src/generated/graphql.ts',
  codegenConfig: {
    scalars: {
      DateTime: 'Date',
    },
  },
});

app.listen({ ...serverConfig }, (err, address) => {
  if (err) {
    console.error(err);
    process.exit(1);
  }
  console.log(`Server listening at ${address} with config:`, config);
});
