import Keyv from 'keyv';
import MemcacheStore from '@keyv/memcache';
import { MemcConfig } from '../config';

/**
 * Service for interacting with Memcached using Keyv as an abstraction layer
 */
export class MemcService {
  private keyv: Keyv<string>;

  /**
   * Creates a new MemcService instance
   * @param config - Memcached configuration including host, port and TTL
   */
  constructor(config: MemcConfig) {
    const memcache = new MemcacheStore(`${config.host}:${config.port}`);
    this.keyv = new Keyv<string>({
      store: memcache as any,
      ttl: config.ttl,
    });
  }

  /**
   * Retrieves a value from Memcached by key
   * @param key - The key to retrieve
   * @returns The value if found, undefined otherwise
   */
  public async get(key: string): Promise<string | undefined> {
    try {
      return await this.keyv.get(key);
    } catch (error) {
      console.error('Error getting value from Memcached:', error);
      return undefined;
    }
  }

  /**
   * Stores a value in Memcached
   * @param key - The key to store the value under
   * @param value - The value to store
   * @param ttl - Optional time-to-live in milliseconds (overrides default TTL)
   */
  public async set(key: string, value: string, ttl?: number): Promise<boolean> {
    try {
      return await this.keyv.set(key, value, ttl);
    } catch (error) {
      console.error('Error setting value in Memcached:', error);
      return false;
    }
  }

  /**
   * Deletes a value from Memcached
   * @param key - The key to delete
   */
  public async delete(key: string): Promise<boolean> {
    try {
      return await this.keyv.delete(key);
    } catch (error) {
      console.error('Error deleting value from Memcached:', error);
      return false;
    }
  }
}
