import jwt from 'jsonwebtoken';
import { v4 as uuidv4 } from 'uuid';
import crypto from 'crypto';
import { TokenConfig } from '../config';
import { PrismaClient } from '@prisma/client';
import { MemcService } from './memc.service';

type TransactionClient = Omit<
  PrismaClient,
  '$connect' | '$disconnect' | '$on' | '$transaction' | '$use' | '$extends'
>;

export interface TokenPayload {
  sub: string;
  role: string;
  deviceId: string;
  jti: string;
}

interface GenerateTokensParams {
  userId: string;
  role: string;
  deviceId: string;
  txn?: TransactionClient;
}

interface RotateRefreshTokenParams {
  userId: string;
  oldToken: string;
  deviceId: string;
}

interface RevokeTokensByDeviceParams {
  userId: string;
  deviceId: string;
}

/**
 * Service for handling token generation, verification, and management
 */
export class TokenService {
  private prisma: PrismaClient;
  private memc: MemcService;
  private config: TokenConfig;

  /**
   * Creates a new TokenService instance
   * @param config - Token configuration
   * @param prisma - Prisma client instance
   * @param memc - Memcached service instance
   */
  constructor({
    prisma,
    memc,
    ...config
  }: TokenConfig & { prisma: PrismaClient; memc: MemcService }) {
    this.prisma = prisma;
    this.memc = memc;
    this.config = config;
  }

  /**
   * Hashes a token using SHA-256
   * @param token - The token to hash
   * @returns Hashed token string
   */
  private async hashToken(token: string): Promise<string> {
    return crypto.createHash('sha256').update(token).digest('hex');
  }

  /**
   * Generates new access and refresh tokens for a user
   * @param params - Object containing userId, role and deviceId
   * @returns Object containing accessToken and refreshToken
   */
  async generateTokens(params: GenerateTokensParams) {
    const { userId, role, deviceId, txn } = params;
    const jti = uuidv4();
    const refreshToken = uuidv4();
    const hash = await this.hashToken(refreshToken);
    const accessToken = jwt.sign(
      {
        sub: userId,
        role,
        deviceId,
        jti,
      },
      this.config.secret,
      { expiresIn: this.config.accessTokenExpiresIn }
    );

    const tx = txn
      ? (callback: Function) => callback(txn)
      : this.prisma.$transaction.bind(this.prisma);

    await tx(async (tx) => {
      await tx.refreshToken.create({
        data: {
          jti,
          userId,
          deviceId,
          hash,
          expiresAt: new Date(Date.now() + this.config.refreshTokenExpiresIn),
        },
      });

      await this.whitelistAccessToken(jti);
    });

    return { accessToken, refreshToken };
  }

  /**
   * Verifies an access token and checks if it's whitelisted
   * @param token - The token to verify
   * @returns Decoded token payload or null if invalid
   */
  async verifyAccessToken(token?: string) {
    try {
      if (!token) return null;
      const decoded = jwt.verify(token, this.config.secret) as TokenPayload;
      const cached = await this.memc.get(`token:whitelist:${decoded.jti}`);
      if (!cached) throw new Error('Token inválido');

      return decoded;
    } catch (error) {
      console.error('Error verifying access token', error);
      return null;
    }
  }

  /**
   * Verifies and rotates a refresh token
   * @param params - Object containing userId, oldToken and deviceId
   * @returns New access and refresh tokens
   */
  async rotateRefreshToken(params: RotateRefreshTokenParams) {
    const { userId, oldToken, deviceId } = params;
    return await this.prisma.$transaction(async (txn) => {
      const storedTokens = await txn.refreshToken.findMany({
        where: { userId, deviceId },
      });
      const hash = await this.hashToken(oldToken);
      const validToken = storedTokens.find((token) => token.hash === hash);
      if (!validToken || validToken.expiresAt < new Date())
        throw new Error('Refresh token inválido');

      await Promise.all(
        storedTokens.map((token) => {
          this.memc.delete(`token:whitelist:${token.jti}`);
        })
      );
      await txn.refreshToken.deleteMany({
        where: { id: { in: storedTokens.map((token) => token.id) } },
      });

      return this.generateTokens({ userId, role: 'user', deviceId, txn });
    });
  }

  /**
   * Revokes all tokens for a user, useful for logout from all devices
   * @param userId - The user ID
   * @returns True if successful
   */
  async revokeAllTokens(userId: string) {
    await this.prisma.$transaction(async (tx) => {
      const tokens = await tx.refreshToken.findMany({
        where: { userId },
      });
      await Promise.all(
        tokens.map((token) => {
          this.memc.delete(`token:whitelist:${token.jti}`);
        })
      );
      await tx.refreshToken.deleteMany({
        where: { id: { in: tokens.map((token) => token.id) } },
      });
    });
    return true;
  }

  /**
   * Revokes tokens for a specific device
   * @param params - Object containing userId and deviceId
   * @returns True if successful
   */
  async revokeTokensByDevice(params: RevokeTokensByDeviceParams) {
    const { userId, deviceId } = params;
    await this.prisma.$transaction(async (tx) => {
      const tokens = await tx.refreshToken.findMany({
        where: { userId, deviceId },
      });
      await Promise.all(
        tokens.map((token) => {
          this.memc.delete(`token:whitelist:${token.jti}`);
        })
      );
      await tx.refreshToken.deleteMany({
        where: { id: { in: tokens.map((token) => token.id) } },
      });
    });
    return true;
  }

  /**
   * Adds an access token to the whitelist
   * @param jti - The JWT ID
   */
  async whitelistAccessToken(jti: string) {
    const ttl = this.config.accessTokenExpiresIn;
    await this.memc.set(`token:whitelist:${jti}`, '1', ttl);
  }

  /**
   * Cleans up expired tokens from the database and Memcached
   */
  async cleanupExpiredTokens() {
    const now = new Date();
    const expired = await this.prisma.refreshToken.findMany({
      where: { expiresAt: { lt: now } },
    });
    await Promise.all(
      expired.map((token) => {
        this.memc.delete(`token:whitelist:${token.jti}`);
      })
    );
    await this.prisma.refreshToken.deleteMany({
      where: { id: { in: expired.map((token) => token.id) } },
    });
  }
}
